import socket

import doctest
import pytest
import requests

import advocate
import advocate.api

from advocate.exceptions import MountDisabledException, ProxyDisabledException
from advocate.packages import ipaddress
from test.monkeypatching import CheckedSocket


SKIP_EXCEPTIONS = (MountDisabledException, ProxyDisabledException)


def pytest_runtestloop():
    validator = advocate.AddrValidator(
        ip_whitelist={
            # requests needs to be able to hit these for its tests!
            ipaddress.ip_network("127.0.0.1"),
            ipaddress.ip_network("127.0.1.1"),
            ipaddress.ip_network("10.255.255.1"),
        },
        # the `httpbin` fixture uses a random port, we need to allow all ports
        port_whitelist=set(range(0, 65535)),
    )

    # this will yell at us if we failed to patch something
    socket.socket = CheckedSocket

    # requests' tests rely on being able to pickle a `Session`
    advocate.api.RequestsAPIWrapper.SUPPORT_WRAPPER_PICKLING = True
    wrapper = advocate.api.RequestsAPIWrapper(validator)

    for attr in advocate.api.__all__:
        setattr(requests, attr, getattr(wrapper, attr))


def pytest_runtest_makereport(item, call):
    # This is necessary because we pull in requests' test suite,
    # which sometimes tests `session.mount()`. We disable that
    # method, so we need to ignore tests that use it.

    from _pytest.runner import pytest_runtest_makereport as mr
    report = mr(item, call)

    if call.excinfo is not None:
        exc = call.excinfo.value
        if isinstance(exc, doctest.UnexpectedException):
            exc = call.excinfo.value.exc_info[1]

        if isinstance(exc, SKIP_EXCEPTIONS):
            report.outcome = 'skipped'
            report.wasxfail = "reason: Advocate is not meant to support this"

    return report



# -*- coding: utf-8 -*-

"""
advocate.api
~~~~~~~~~~~~

This module implements the Requests API, largely a copy/paste from `requests`
itself.

:copyright: (c) 2015 by Jordan Milne.
:license: Apache2, see LICENSE for more details.

"""
from collections import OrderedDict
import hashlib
import pickle

from requests import Session as RequestsSession

import advocate
from .adapters import ValidatingHTTPAdapter
from .exceptions import MountDisabledException


class Session(RequestsSession):
    __attrs__ = RequestsSession.__attrs__ + ["validator"]
    DEFAULT_VALIDATOR = None

    """Convenience wrapper around `requests.Session` set up for `advocate`ing"""
    def __init__(self, *args, **kwargs):
        self.validator = kwargs.pop("validator", self.DEFAULT_VALIDATOR)
        adapter_kwargs = kwargs.pop("_adapter_kwargs", {})

        # `Session.__init__()` calls `mount()` internally, so we need to allow
        # it temporarily
        self.__mountAllowed = True
        RequestsSession.__init__(self, *args, **kwargs)

        # Drop any existing adapters
        self.adapters = OrderedDict()

        self.mount("http://", ValidatingHTTPAdapter(validator=self.validator, **adapter_kwargs))
        self.mount("https://", ValidatingHTTPAdapter(validator=self.validator, **adapter_kwargs))
        self.__mountAllowed = False

    def mount(self, *args, **kwargs):
        """Wrapper around `mount()` to prevent a protection bypass"""
        if self.__mountAllowed:
            super(Session, self).mount(*args, **kwargs)
        else:
            raise MountDisabledException(
                "mount() is disabled to prevent protection bypasses"
            )


def session(*args, **kwargs):
    return Session(*args, **kwargs)


def request(method, url, **kwargs):
    """Constructs and sends a :class:`Request <Request>`.

    :param method: method for the new :class:`Request` object.
    :param url: URL for the new :class:`Request` object.
    :param params: (optional) Dictionary or bytes to be sent in the query string for the :class:`Request`.
    :param data: (optional) Dictionary, bytes, or file-like object to send in the body of the :class:`Request`.
    :param json: (optional) json data to send in the body of the :class:`Request`.
    :param headers: (optional) Dictionary of HTTP Headers to send with the :class:`Request`.
    :param cookies: (optional) Dict or CookieJar object to send with the :class:`Request`.
    :param files: (optional) Dictionary of ``'name': file-like-objects`` (or ``{'name': ('filename', fileobj)}``) for multipart encoding upload.
    :param auth: (optional) Auth tuple to enable Basic/Digest/Custom HTTP Auth.
    :param timeout: (optional) How long to wait for the server to send data
        before giving up, as a float, or a (`connect timeout, read timeout
        <user/advanced.html#timeouts>`_) tuple.
    :type timeout: float or tuple
    :param allow_redirects: (optional) Boolean. Set to True if POST/PUT/DELETE redirect following is allowed.
    :type allow_redirects: bool
    :param proxies: (optional) Dictionary mapping protocol to the URL of the proxy.
    :param verify: (optional) if ``True``, the SSL cert will be verified. A CA_BUNDLE path can also be provided.
    :param stream: (optional) if ``False``, the response content will be immediately downloaded.
    :param cert: (optional) if String, path to ssl client cert file (.pem). If Tuple, ('cert', 'key') pair.
    :return: :class:`Response <Response>` object
    :rtype: requests.Response

    Usage::

      >>> import advocate
      >>> req = advocate.request('GET', 'http://httpbin.org/get')
      <Response [200]>
    """

    validator = kwargs.pop("validator", None)
    with Session(validator=validator) as sess:
        response = sess.request(method=method, url=url, **kwargs)
    return response


def get(url, **kwargs):
    """Sends a GET request.

    :param url: URL for the new :class:`Request` object.
    :param \*\*kwargs: Optional arguments that ``request`` takes.
    :return: :class:`Response <Response>` object
    :rtype: requests.Response
    """

    kwargs.setdefault('allow_redirects', True)
    return request('get', url, **kwargs)


def options(url, **kwargs):
    """Sends a OPTIONS request.

    :param url: URL for the new :class:`Request` object.
    :param \*\*kwargs: Optional arguments that ``request`` takes.
    :return: :class:`Response <Response>` object
    :rtype: requests.Response
    """

    kwargs.setdefault('allow_redirects', True)
    return request('options', url, **kwargs)


def head(url, **kwargs):
    """Sends a HEAD request.

    :param url: URL for the new :class:`Request` object.
    :param \*\*kwargs: Optional arguments that ``request`` takes.
    :return: :class:`Response <Response>` object
    :rtype: requests.Response
    """

    kwargs.setdefault('allow_redirects', False)
    return request('head', url, **kwargs)


def post(url, data=None, json=None, **kwargs):
    """Sends a POST request.

    :param url: URL for the new :class:`Request` object.
    :param data: (optional) Dictionary, bytes, or file-like object to send in the body of the :class:`Request`.
    :param json: (optional) json data to send in the body of the :class:`Request`.
    :param \*\*kwargs: Optional arguments that ``request`` takes.
    :return: :class:`Response <Response>` object
    :rtype: requests.Response
    """

    return request('post', url, data=data, json=json, **kwargs)


def put(url, data=None, **kwargs):
    """Sends a PUT request.

    :param url: URL for the new :class:`Request` object.
    :param data: (optional) Dictionary, bytes, or file-like object to send in the body of the :class:`Request`.
    :param \*\*kwargs: Optional arguments that ``request`` takes.
    :return: :class:`Response <Response>` object
    :rtype: requests.Response
    """

    return request('put', url, data=data, **kwargs)


def patch(url, data=None, **kwargs):
    """Sends a PATCH request.

    :param url: URL for the new :class:`Request` object.
    :param data: (optional) Dictionary, bytes, or file-like object to send in the body of the :class:`Request`.
    :param \*\*kwargs: Optional arguments that ``request`` takes.
    :return: :class:`Response <Response>` object
    :rtype: requests.Response
    """

    return request('patch', url, data=data, **kwargs)


def delete(url, **kwargs):
    """Sends a DELETE request.

    :param url: URL for the new :class:`Request` object.
    :param \*\*kwargs: Optional arguments that ``request`` takes.
    :return: :class:`Response <Response>` object
    :rtype: requests.Response
    """

    return request('delete', url, **kwargs)


class RequestsAPIWrapper(object):
    """Provides a `requests.api`-like interface with a specific validator"""

    # Due to how the classes are dynamically constructed pickling may not work
    # correctly unless loaded within the same interpreter instance.
    # Enable at your peril.
    SUPPORT_WRAPPER_PICKLING = False

    def __init__(self, validator):
        # Do this here to avoid circular import issues
        try:
            from .futures import FuturesSession
            have_requests_futures = True
        except ImportError as e:
            have_requests_futures = False

        self.validator = validator
        outer_self = self

        class _WrappedSession(Session):
            """An `advocate.Session` that uses the wrapper's blacklist

            the wrapper is meant to be a transparent replacement for `requests`,
            so people should be able to subclass `wrapper.Session` and still
            get the desired validation behaviour
            """
            DEFAULT_VALIDATOR = outer_self.validator

        self._make_wrapper_cls_global(_WrappedSession)

        if have_requests_futures:

            class _WrappedFuturesSession(FuturesSession):
                """Like _WrappedSession, but for `FuturesSession`s"""
                DEFAULT_VALIDATOR = outer_self.validator
            self._make_wrapper_cls_global(_WrappedFuturesSession)

            self.FuturesSession = _WrappedFuturesSession

        self.request = self._default_arg_wrapper(request)
        self.get = self._default_arg_wrapper(get)
        self.options = self._default_arg_wrapper(options)
        self.head = self._default_arg_wrapper(head)
        self.post = self._default_arg_wrapper(post)
        self.put = self._default_arg_wrapper(put)
        self.patch = self._default_arg_wrapper(patch)
        self.delete = self._default_arg_wrapper(delete)
        self.session = self._default_arg_wrapper(session)
        self.Session = _WrappedSession

    def __getattr__(self, item):
        # This class is meant to mimic the requests base module, so if we don't
        # have this attribute, it might be on the base module (like the Request
        # class, etc.)
        try:
            return object.__getattribute__(self, item)
        except AttributeError:
            return getattr(advocate, item)

    def _default_arg_wrapper(self, fun):
        def wrapped_func(*args, **kwargs):
            kwargs.setdefault("validator", self.validator)
            return fun(*args, **kwargs)
        return wrapped_func

    def _make_wrapper_cls_global(self, cls):
        if not self.SUPPORT_WRAPPER_PICKLING:
            return
        # Gnarly, but necessary to give pickle a consistent module-level
        # reference for each wrapper.
        wrapper_hash = hashlib.sha256(pickle.dumps(self)).hexdigest()
        cls.__name__ = "_".join((cls.__name__, wrapper_hash))
        cls.__qualname__ = ".".join((__name__, cls.__name__))
        if not globals().get(cls.__name__):
            globals()[cls.__name__] = cls


__all__ = (
    "delete",
    "get",
    "head",
    "options",
    "patch",
    "post",
    "put",
    "request",
    "session",
    "Session",
    "RequestsAPIWrapper",
)



from __future__ import absolute_import

# XXX: Python 3 before 3.3 doesn't have the `ipaddress` module and will
# break on this import.
try:
    # First try to use our bundled ipaddress module
    from .ipaddress import *
except ImportError:
    # Try to pull from the global `ipaddress` module
    from ipaddress import *



class AdvocateException(Exception):
    pass


class UnacceptableAddressException(AdvocateException):
    pass


class NameserverException(AdvocateException):
    pass


class MountDisabledException(AdvocateException):
    pass


class ProxyDisabledException(NotImplementedError, AdvocateException):
    pass



__version__ = "1.0.0"

from requests import utils
from requests.models import Request, Response, PreparedRequest
from requests.status_codes import codes
from requests.exceptions import (
    RequestException, Timeout, URLRequired,
    TooManyRedirects, HTTPError, ConnectionError
)

from .adapters import ValidatingHTTPAdapter
from .api import *
from .addrvalidator import AddrValidator
from .exceptions import UnacceptableAddressException



from requests.packages.urllib3 import HTTPConnectionPool, HTTPSConnectionPool

from .connection import (
    ValidatingHTTPConnection,
    ValidatingHTTPSConnection,
)

# Don't silently break if the private API changes across urllib3 versions
assert(hasattr(HTTPConnectionPool, 'ConnectionCls'))
assert(hasattr(HTTPSConnectionPool, 'ConnectionCls'))
assert(hasattr(HTTPConnectionPool, 'scheme'))
assert(hasattr(HTTPSConnectionPool, 'scheme'))


class ValidatingHTTPConnectionPool(HTTPConnectionPool):
    scheme = 'http'
    ConnectionCls = ValidatingHTTPConnection


class ValidatingHTTPSConnectionPool(HTTPSConnectionPool):
    scheme = 'https'
    ConnectionCls = ValidatingHTTPSConnection



# coding=utf-8

from __future__ import division

import pickle
import socket
import unittest

# This needs to be done before third-party imports to make sure they all use
# our wrapped socket class, especially in case of subclasses.
from .monkeypatching import CheckedSocket, DisallowedConnectException
socket.socket = CheckedSocket

from mock import patch
import requests
import requests_mock
import six.moves

import advocate
from advocate import AddrValidator
from advocate.addrvalidator import canonicalize_hostname
from advocate.api import RequestsAPIWrapper
from advocate.connection import advocate_getaddrinfo
from advocate.exceptions import (
    MountDisabledException,
    NameserverException,
    UnacceptableAddressException,
)
from advocate.packages import ipaddress
from advocate.futures import FuturesSession


# We use port 1 for testing because nothing is likely to legitimately listen
# on it.
AddrValidator.DEFAULT_PORT_WHITELIST.add(1)

RequestsAPIWrapper.SUPPORT_WRAPPER_PICKLING = True
global_wrapper = RequestsAPIWrapper(validator=AddrValidator(ip_whitelist={
    ipaddress.ip_network("127.0.0.1"),
}))
RequestsAPIWrapper.SUPPORT_WRAPPER_PICKLING = False


class _WrapperSubclass(global_wrapper.Session):
    def good_method(self):
        return "foo"


def canonname_supported():
    """Check if the nameserver supports the AI_CANONNAME flag

    travis-ci.org's Python 3 env doesn't seem to support it, so don't try
    any of the test that rely on it.
    """
    addrinfo = advocate_getaddrinfo("example.com", 0, get_canonname=True)
    assert addrinfo
    return addrinfo[0][3] == b"example.com"


def permissive_validator(**kwargs):
    default_options = dict(
        ip_blacklist=None,
        port_whitelist=None,
        port_blacklist=None,
        hostname_blacklist=None,
        allow_ipv6=True,
        allow_teredo=True,
        allow_6to4=True,
        allow_dns64=True,
        autodetect_local_addresses=False,
    )
    default_options.update(**kwargs)
    return AddrValidator(**default_options)


# Test our test wrappers to make sure they're testy
class TestWrapperTests(unittest.TestCase):
    def test_unsafe_connect_raises(self):
        self.assertRaises(
            DisallowedConnectException,
            requests.get, "http://example.org/"
        )


class ValidateIPTests(unittest.TestCase):
    def _test_ip_kind_blocked(self, ip, **kwargs):
        validator = permissive_validator(**kwargs)
        self.assertFalse(validator.is_ip_allowed(ip))

    def test_manual_ip_blacklist(self):
        """Test manually blacklisting based on IP"""
        validator = AddrValidator(
            allow_ipv6=True,
            ip_blacklist=(
                ipaddress.ip_network("132.0.5.0/24"),
                ipaddress.ip_network("152.0.0.0/8"),
                ipaddress.ip_network("::1"),
            ),
        )
        self.assertFalse(validator.is_ip_allowed("132.0.5.1"))
        self.assertFalse(validator.is_ip_allowed("152.254.90.1"))
        self.assertTrue(validator.is_ip_allowed("178.254.90.1"))
        self.assertFalse(validator.is_ip_allowed("::1"))
        # Google, found via `dig google.com AAAA`
        self.assertTrue(validator.is_ip_allowed("2607:f8b0:400a:807::200e"))

    def test_ip_whitelist(self):
        """Test manually whitelisting based on IP"""
        validator = AddrValidator(
            ip_whitelist=(
                ipaddress.ip_network("127.0.0.1"),
            ),
        )
        self.assertTrue(validator.is_ip_allowed("127.0.0.1"))

    def test_ip_whitelist_blacklist_conflict(self):
        """Manual whitelist should take precedence over manual blacklist"""
        validator = AddrValidator(
            ip_whitelist=(
                ipaddress.ip_network("127.0.0.1"),
            ),
            ip_blacklist=(
                ipaddress.ip_network("127.0.0.1"),
            ),
        )
        self.assertTrue(validator.is_ip_allowed("127.0.0.1"))

    @unittest.skip("takes half an hour or so to run")
    def test_safecurl_blacklist(self):
        """Test that we at least disallow everything SafeCurl does"""
        # All IPs that SafeCurl would disallow
        bad_netblocks = (ipaddress.ip_network(x) for x in (
            '0.0.0.0/8',
            '10.0.0.0/8',
            '100.64.0.0/10',
            '127.0.0.0/8',
            '169.254.0.0/16',
            '172.16.0.0/12',
            '192.0.0.0/29',
            '192.0.2.0/24',
            '192.88.99.0/24',
            '192.168.0.0/16',
            '198.18.0.0/15',
            '198.51.100.0/24',
            '203.0.113.0/24',
            '224.0.0.0/4',
            '240.0.0.0/4'
        ))
        i = 0
        validator = AddrValidator()
        for bad_netblock in bad_netblocks:
            num_ips = bad_netblock.num_addresses
            # Don't test *every* IP in large netblocks
            step_size = int(min(max(num_ips / 255, 1), 128))
            for ip_idx in six.moves.range(0, num_ips, step_size):
                i += 1
                bad_ip = bad_netblock[ip_idx]
                bad_ip_allowed = validator.is_ip_allowed(bad_ip)
                if bad_ip_allowed:
                    print(i, bad_ip)
                self.assertFalse(bad_ip_allowed)

    # TODO: something like the above for IPv6?

    def test_ipv4_mapped(self):
        self._test_ip_kind_blocked("::ffff:192.168.2.1")

    def test_teredo(self):
        # 192.168.2.1 as the client address
        self._test_ip_kind_blocked("2001:0000:4136:e378:8000:63bf:3f57:fdf2")
        # This should be disallowed even if teredo is allowed.
        self._test_ip_kind_blocked(
            "2001:0000:4136:e378:8000:63bf:3f57:fdf2",
            allow_teredo=False,
        )

    def test_ipv6(self):
        self._test_ip_kind_blocked("2002:C0A8:FFFF::", allow_ipv6=False)

    def test_sixtofour(self):
        # 192.168.XXX.XXX
        self._test_ip_kind_blocked("2002:C0A8:FFFF::")
        self._test_ip_kind_blocked("2002:C0A8:FFFF::", allow_6to4=False)

    def test_dns64(self):
        # XXX: Don't even know if this is an issue, TBH. Seems to be related
        # to DNS64/NAT64, but not a lot of easy-to-understand info:
        # https://tools.ietf.org/html/rfc6052
        self._test_ip_kind_blocked("64:ff9b::192.168.2.1")
        self._test_ip_kind_blocked("64:ff9b::192.168.2.1", allow_dns64=False)

    def test_link_local(self):
        # 169.254.XXX.XXX, AWS uses these for autoconfiguration
        self._test_ip_kind_blocked("169.254.1.1")

    def test_site_local(self):
        self._test_ip_kind_blocked("FEC0:CCCC::")

    def test_loopback(self):
        self._test_ip_kind_blocked("127.0.0.1")
        self._test_ip_kind_blocked("::1")

    def test_multicast(self):
        self._test_ip_kind_blocked("227.1.1.1")

    def test_private(self):
        self._test_ip_kind_blocked("192.168.2.1")
        self._test_ip_kind_blocked("10.5.5.5")
        self._test_ip_kind_blocked("0.0.0.0")
        self._test_ip_kind_blocked("0.1.1.1")
        self._test_ip_kind_blocked("100.64.0.0")

    def test_reserved(self):
        self._test_ip_kind_blocked("255.255.255.255")
        self._test_ip_kind_blocked("::ffff:192.168.2.1")
        # 6to4 relay
        self._test_ip_kind_blocked("192.88.99.0")

    def test_unspecified(self):
        self._test_ip_kind_blocked("0.0.0.0")

    def test_parsed(self):
        validator = permissive_validator()
        self.assertFalse(validator.is_ip_allowed(
            ipaddress.ip_address("0.0.0.0")
        ))
        self.assertTrue(validator.is_ip_allowed(
            ipaddress.ip_address("144.1.1.1")
        ))


class AddrInfoTests(unittest.TestCase):
    def _is_addrinfo_allowed(self, host, port, **kwargs):
        validator = permissive_validator(**kwargs)
        allowed = False
        for res in advocate_getaddrinfo(host, port):
            if validator.is_addrinfo_allowed(res):
                allowed = True
        return allowed

    def test_simple(self):
        self.assertFalse(
            self._is_addrinfo_allowed("192.168.0.1", 80)
        )

    def test_malformed_addrinfo(self):
        # Alright, the addrinfo format is probably never going to change,
        # but *what if it did?*
        vl = permissive_validator()
        addrinfo = advocate_getaddrinfo("example.com", 80)[0] + (1,)
        self.assertRaises(Exception, lambda: vl.is_addrinfo_allowed(addrinfo))

    def test_unexpected_proto(self):
        # What if addrinfo returns info about a protocol we don't understand?
        vl = permissive_validator()
        addrinfo = list(advocate_getaddrinfo("example.com", 80)[0])
        addrinfo[4] = addrinfo[4] + (1,)
        self.assertRaises(Exception, lambda: vl.is_addrinfo_allowed(addrinfo))

    def test_default_port_whitelist(self):
        self.assertTrue(
            self._is_addrinfo_allowed("200.1.1.1", 8080)
        )
        self.assertTrue(
            self._is_addrinfo_allowed("200.1.1.1", 80)
        )
        self.assertFalse(
            self._is_addrinfo_allowed("200.1.1.1", 99)
        )

    def test_port_whitelist(self):
        wl = (80, 10)
        self.assertTrue(
            self._is_addrinfo_allowed("200.1.1.1", 80, port_whitelist=wl)
        )
        self.assertTrue(
            self._is_addrinfo_allowed("200.1.1.1", 10, port_whitelist=wl)
        )
        self.assertFalse(
            self._is_addrinfo_allowed("200.1.1.1", 99, port_whitelist=wl)
        )

    def test_port_blacklist(self):
        bl = (80, 10)
        self.assertFalse(
            self._is_addrinfo_allowed("200.1.1.1", 80, port_blacklist=bl)
        )
        self.assertFalse(
            self._is_addrinfo_allowed("200.1.1.1", 10, port_blacklist=bl)
        )
        self.assertTrue(
            self._is_addrinfo_allowed("200.1.1.1", 99, port_blacklist=bl)
        )

    @patch("advocate.addrvalidator.determine_local_addresses")
    def test_local_address_handling(self, mock_determine_local_addresses):
        fake_addresses = [ipaddress.ip_network("200.1.1.1")]
        mock_determine_local_addresses.return_value = fake_addresses

        self.assertFalse(self._is_addrinfo_allowed(
            "200.1.1.1",
            80,
            autodetect_local_addresses=True
        ))
        # Check that `is_ip_allowed` didn't make its own call to determine
        # local addresses
        mock_determine_local_addresses.assert_called_once_with()
        mock_determine_local_addresses.reset_mock()

        self.assertTrue(self._is_addrinfo_allowed(
            "200.1.1.1",
            80,
            autodetect_local_addresses=False,
        ))
        mock_determine_local_addresses.assert_not_called()


@unittest.skipIf(
    not canonname_supported(),
    "Nameserver doesn't support AI_CANONNAME, skipping hostname tests"
)
class HostnameTests(unittest.TestCase):
    def setUp(self):
        self._canonname_supported = canonname_supported()

    def _is_hostname_allowed(self, host, fake_lookup=False, **kwargs):
        validator = permissive_validator(**kwargs)
        if fake_lookup:
            results = [(2, 1, 6, canonicalize_hostname(host).encode("utf8"), ('1.2.3.4', 80))]
        else:
            results = advocate_getaddrinfo(host, 80, get_canonname=True)
        for res in results:
            if validator.is_addrinfo_allowed(res):
                return True
        return False

    def test_no_blacklist(self):
        self.assertTrue(self._is_hostname_allowed("example.com"))

    def test_idn(self):
        # test some basic globs
        self.assertFalse(self._is_hostname_allowed(
            u"中国.example.org",
            fake_lookup=True,
            hostname_blacklist={"*.org"}
        ))
        # case insensitive, please
        self.assertFalse(self._is_hostname_allowed(
            u"中国.example.oRg",
            fake_lookup=True,
            hostname_blacklist={"*.Org"}
        ))
        self.assertFalse(self._is_hostname_allowed(
            u"中国.example.org",
            fake_lookup=True,
            hostname_blacklist={"xn--fiqs8s.*.org"}
        ))
        self.assertFalse(self._is_hostname_allowed(
            "xn--fiqs8s.example.org",
            fake_lookup=True,
            hostname_blacklist={u"中国.*.org"}
        ))
        self.assertTrue(self._is_hostname_allowed(
            u"example.org",
            fake_lookup=True,
            hostname_blacklist={u"中国.*.org"}
        ))
        self.assertTrue(self._is_hostname_allowed(
            u"example.com",
            fake_lookup=True,
            hostname_blacklist={u"中国.*.org"}
        ))
        self.assertTrue(self._is_hostname_allowed(
            u"foo.example.org",
            fake_lookup=True,
            hostname_blacklist={u"中国.*.org"}
        ))

    def test_missing_canonname(self):
        addrinfo = socket.getaddrinfo(
            "127.0.0.1",
            1,
            0,
            socket.SOCK_STREAM,
        )
        self.assertTrue(addrinfo)

        # Should throw an error if we're using hostname blacklisting and the
        # addrinfo record we passed in doesn't have a canonname
        validator = permissive_validator(hostname_blacklist={"foo"})
        self.assertRaises(
            NameserverException,
            validator.is_addrinfo_allowed, addrinfo[0]
        )

    def test_embedded_null(self):
        vl = permissive_validator(hostname_blacklist={"*.baz.com"})
        # Things get a little screwy with embedded nulls. Try to emulate any
        # possible null termination when checking if the hostname is allowed.
        self.assertFalse(vl.is_hostname_allowed("foo.baz.com\x00.example.com"))
        self.assertFalse(vl.is_hostname_allowed("foo.example.com\x00.baz.com"))
        self.assertFalse(vl.is_hostname_allowed(u"foo.baz.com\x00.example.com"))
        self.assertFalse(vl.is_hostname_allowed(u"foo.example.com\x00.baz.com"))


class ConnectionPoolingTests(unittest.TestCase):
    @patch("advocate.connection.ValidatingHTTPConnection._new_conn")
    def test_connection_reuse(self, mock_new_conn):
        # Just because you can use an existing connection doesn't mean you
        # should. The disadvantage of us working at the socket level means that
        # we get bitten if a connection pool is shared between regular requests
        # and advocate.
        # This can never happen with requests, but let's set a good example :)
        with CheckedSocket.bypass_checks():
            # HTTPBin supports `keep-alive`, so it's a good test subject
            requests.get("http://httpbin.org/")
        try:
            advocate.get("http://httpbin.org/")
        except:
            pass
        # Requests may retry several times, but our mock doesn't return a real
        # socket. Just check that it tried to create one.
        mock_new_conn.assert_any_call()


class AdvocateWrapperTests(unittest.TestCase):
    def test_get(self):
        self.assertEqual(advocate.get("http://example.com").status_code, 200)
        self.assertEqual(advocate.get("https://example.com").status_code, 200)

    def test_validator(self):
        self.assertRaises(
            UnacceptableAddressException,
            advocate.get, "http://127.0.0.1/"
        )
        self.assertRaises(
            UnacceptableAddressException,
            advocate.get, "http://localhost/"
        )
        self.assertRaises(
            UnacceptableAddressException,
            advocate.get, "https://localhost/"
        )

    @unittest.skipIf(
        not canonname_supported(),
        "Nameserver doesn't support AI_CANONNAME, skipping hostname tests"
    )
    def test_blacklist_hostname(self):
        self.assertRaises(
            UnacceptableAddressException,
            advocate.get,
            "https://google.com/",
            validator=AddrValidator(hostname_blacklist={"google.com"})
        )

    # Disabled for now because the redirection endpoint appears to be broken.
    @unittest.skip
    def test_redirect(self):
        # Make sure httpbin even works
        test_url = "http://httpbin.org/status/204"
        self.assertEqual(advocate.get(test_url).status_code, 204)

        redir_url = "http://httpbin.org/redirect-to?url=http://127.0.0.1/"
        self.assertRaises(
            UnacceptableAddressException,
            advocate.get, redir_url
        )

    def test_mount_disabled(self):
        sess = advocate.Session()
        self.assertRaises(
            MountDisabledException,
            sess.mount,
            "foo://",
            None,
        )

    def test_advocate_requests_api_wrapper(self):
        wrapper = RequestsAPIWrapper(validator=AddrValidator())
        local_validator = AddrValidator(ip_whitelist={
            ipaddress.ip_network("127.0.0.1"),
        })
        local_wrapper = RequestsAPIWrapper(validator=local_validator)

        self.assertRaises(
            UnacceptableAddressException,
            wrapper.get, "http://127.0.0.1:1/"
        )

        with self.assertRaises(Exception) as cm:
            local_wrapper.get("http://127.0.0.1:1/")
        # Check that we got a connection exception instead of a validation one
        # This might be either exception depending on the requests version
        self.assertRegexpMatches(
            cm.exception.__class__.__name__,
            r"\A(Connection|Protocol)Error",
        )
        self.assertRaises(
            UnacceptableAddressException,
            wrapper.get, "http://localhost:1/"
        )
        self.assertRaises(
            UnacceptableAddressException,
            wrapper.get, "https://localhost:1/"
        )

    def test_wrapper_session_pickle(self):
        # Make sure the validator still works after a pickle round-trip
        sess_instance = pickle.loads(pickle.dumps(global_wrapper.Session()))

        with self.assertRaises(Exception) as cm:
            sess_instance.get("http://127.0.0.1:1/")
        self.assertRegexpMatches(
            cm.exception.__class__.__name__,
            r"\A(Connection|Protocol)Error",
        )
        self.assertRaises(
            UnacceptableAddressException,
            sess_instance.get, "http://127.0.1.1:1/"
        )

    def test_wrapper_session_subclass(self):
        # Make sure pickle doesn't explode if we try to pickle a subclass
        # of `global_wrapper.Session`
        def _check_instance(instance):
            self.assertEqual(instance.good_method(), "foo")

            with self.assertRaises(Exception) as cm:
                instance.get("http://127.0.0.1:1/")
            self.assertRegexpMatches(
                cm.exception.__class__.__name__,
                r"\A(Connection|Protocol)Error",
            )
            self.assertRaises(
                UnacceptableAddressException,
                instance.get, "http://127.0.1.1:1/"
            )
        sess = _WrapperSubclass()
        _check_instance(sess)
        sess_unpickled = pickle.loads(pickle.dumps(sess))
        _check_instance(sess_unpickled)



    @unittest.skipIf(
        not canonname_supported(),
        "Nameserver doesn't support AI_CANONNAME, skipping hostname tests"
    )
    def test_advocate_requests_api_wrapper_hostnames(self):
        wrapper = RequestsAPIWrapper(validator=AddrValidator(
            hostname_blacklist={"google.com"},
        ))
        self.assertRaises(
            UnacceptableAddressException,
            wrapper.get,
            "https://google.com/",
        )

    def test_advocate_requests_api_wrapper_req_methods(self):
        # Make sure all the convenience methods make requests with the correct
        # methods
        wrapper = RequestsAPIWrapper(AddrValidator())

        request_methods = (
            "get", "options", "head", "post", "put", "patch", "delete"
        )
        for method_name in request_methods:
            with requests_mock.mock() as request_mock:
                # This will fail if the request expected by `request_mock`
                # isn't sent when calling the wrapper method
                request_mock.request(method_name, "http://example.com/foo")
                getattr(wrapper, method_name)("http://example.com/foo")

    def test_wrapper_getattr_fallback(self):
        # Make sure wrappers include everything in Advocate's `__init__.py`
        wrapper = RequestsAPIWrapper(AddrValidator())
        self.assertIsNotNone(wrapper.PreparedRequest)

    def test_proxy_attempt_throws(self):
        # Advocate can't do anything useful when you use a proxy, the proxy
        # is the one that ultimately makes the connection
        self.assertRaises(
            NotImplementedError,
            advocate.get, "http://example.org/",
            proxies={
                "http": "http://example.org:3128",
                "https": "http://example.org:1080",
            },
        )

    @patch("advocate.addrvalidator.determine_local_addresses")
    def test_connect_without_local_addresses(self, mock_determine_local_addresses):
        fake_addresses = [ipaddress.ip_network("200.1.1.1")]
        mock_determine_local_addresses.return_value = fake_addresses

        validator = permissive_validator(autodetect_local_addresses=True)
        advocate.get("http://example.com/", validator=validator)
        # Check that `is_ip_allowed` didn't make its own call to determine
        # local addresses
        mock_determine_local_addresses.assert_called_once_with()
        mock_determine_local_addresses.reset_mock()

        validator = permissive_validator(autodetect_local_addresses=False)
        advocate.get("http://example.com", validator=validator)
        mock_determine_local_addresses.assert_not_called()


class AdvocateFuturesTest(unittest.TestCase):
    def test_get(self):
        sess = FuturesSession()
        assert 200 == sess.get("http://example.org/").result().status_code

    def test_custom_validator(self):
        validator = AddrValidator(hostname_blacklist={"example.org"})
        sess = FuturesSession(validator=validator)
        self.assertRaises(
            UnacceptableAddressException,
            lambda: sess.get("http://example.org").result()
        )

    def test_many_workers(self):
        sess = FuturesSession(max_workers=50)
        self.assertRaises(
            UnacceptableAddressException,
            lambda: sess.get("http://127.0.0.1:1/").result()
        )

    def test_passing_session(self):
        try:
            FuturesSession(session=requests.Session())
            assert False
        except NotImplementedError:
            pass

        sess = FuturesSession()
        try:
            sess.session = requests.Session()
            assert False
        except NotImplementedError:
            pass

        sess.session = advocate.Session()

    def test_advocate_wrapper_futures(self):
        wrapper = RequestsAPIWrapper(validator=AddrValidator())
        local_validator = AddrValidator(ip_whitelist={
            ipaddress.ip_network("127.0.0.1"),
        })
        local_wrapper = RequestsAPIWrapper(validator=local_validator)

        with self.assertRaises(UnacceptableAddressException):
            sess = wrapper.FuturesSession()
            sess.get("http://127.0.0.1/").result()

        with self.assertRaises(Exception) as cm:
            sess = local_wrapper.FuturesSession()
            sess.get("http://127.0.0.1:1/").result()
        # Check that we got a connection exception instead of a validation one
        # This might be either exception depending on the requests version
        self.assertRegexpMatches(
            cm.exception.__class__.__name__,
            r"\A(Connection|Protocol)Error",
        )

        with self.assertRaises(UnacceptableAddressException):
            sess = wrapper.FuturesSession()
            sess.get("http://localhost:1/").result()
        with self.assertRaises(UnacceptableAddressException):
            sess = wrapper.FuturesSession()
            sess.get("https://localhost:1/").result()


if __name__ == '__main__':
    unittest.main()



import hashlib

from flask import Flask, request
import advocate
import requests

app = Flask(__name__)


@app.route('/')
def get_hash():
    url = request.args.get("url")
    if not url:
        return "Please specify a url!"
    try:
        headers = {"User-Agent": "Hashifier 0.1"}
        resp = advocate.get(url, headers=headers)
    except advocate.UnacceptableAddressException:
        return "That URL points to a forbidden resource"
    except requests.RequestException:
        return "Failed to connect to the specified URL"

    return hashlib.sha256(resp.content).hexdigest()

if __name__ == '__main__':
    app.run()



import contextlib
import os.path
import socket
import traceback


class DisallowedConnectException(Exception):
    pass


class CheckedSocket(socket.socket):
    CONNECT_ALLOWED_FUNCS = {"validating_create_connection"}
    # `test_testserver.py` makes raw connections to the test server to ensure it works
    CONNECT_ALLOWED_FILES = {"test_testserver.py"}
    _checks_enabled = True

    @classmethod
    @contextlib.contextmanager
    def bypass_checks(cls):
        try:
            cls._checks_enabled = False
            yield
        finally:
            cls._checks_enabled = True

    @classmethod
    def _check_frame_allowed(cls, frame):
        if os.path.basename(frame[0]) in cls.CONNECT_ALLOWED_FILES:
            return True
        if frame[2] in cls.CONNECT_ALLOWED_FUNCS:
            return True
        return False

    def connect(self, *args, **kwargs):
        if self._checks_enabled:

            stack = traceback.extract_stack()
            if not any(self._check_frame_allowed(frame) for frame in stack):
                raise DisallowedConnectException("calling socket.connect() unsafely!")
        return super(CheckedSocket, self).connect(*args, **kwargs)



import re
import setuptools
from codecs import open

requires = [
    'requests <3.0, >=2.18.0',
    'urllib3 <2.0, >=1.22',
    'six',
    "pyasn1",
    "pyopenssl",
    "ndg-httpsclient",
    'netifaces>=0.10.5',
]

packages = [
    "advocate",
    "advocate.packages",
    "advocate.packages.ipaddress"
]

version = ''
with open('advocate/__init__.py', 'r') as fd:
    version = re.search(r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]',
                        fd.read(), re.MULTILINE).group(1)

with open('README.rst', 'r', 'utf-8') as f:
    readme = f.read()


setuptools.setup(
    name='advocate',
    version=version,
    packages=packages,
    install_requires=requires,
    tests_require=[
        "mock",
        "pytest",
        "pytest-cov",
        "requests-futures",
        "requests-mock",
    ],
    url='https://github.com/JordanMilne/Advocate',
    license='Apache 2',
    author='Jordan Milne',
    author_email='advocate@saynotolinux.com',
    keywords="http requests security ssrf proxy rebinding advocate",
    description=('A wrapper around the requests library for safely '
                 'making HTTP requests on behalf of a third party'),
    long_description=readme,
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Topic :: Security',
        'Topic :: Internet :: WWW/HTTP',
    ],
)



import requests_futures.sessions
from concurrent.futures import ThreadPoolExecutor
from requests.adapters import DEFAULT_POOLSIZE

from . import Session


class FuturesSession(requests_futures.sessions.FuturesSession, Session):
    def __init__(self, executor=None, max_workers=2, session=None, *args,
                 **kwargs):
        adapter_kwargs = {}
        if executor is None:
            executor = ThreadPoolExecutor(max_workers=max_workers)
            # set connection pool size equal to max_workers if needed
            if max_workers > DEFAULT_POOLSIZE:
                adapter_kwargs = dict(pool_connections=max_workers,
                                      pool_maxsize=max_workers)
        kwargs["_adapter_kwargs"] = adapter_kwargs
        Session.__init__(self, *args, **kwargs)
        self.executor = executor
        self.session = session

    @property
    def session(self):
        return None

    @session.setter
    def session(self, value):
        if value is not None and not isinstance(value, Session):
            raise NotImplementedError("Setting the .session property to "
                                      "non-advocate values disabled "
                                      "to prevent whitelist bypasses")
